/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore
National Laboratory.
*/

#include <pbs_config.h>   /* the master config generated by configure */

#include	<assert.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>
#include	<ctype.h>
#include	<errno.h>
#include	<time.h>
#include	<limits.h>
#include	<netdb.h>
#include	<sys/types.h>
#include	<sys/param.h>
#include	<sys/times.h>
#include	<sys/stat.h>
#include	<netinet/in.h>
#include	<sys/time.h>

#include 	"pbs_ifl.h"
#include	"pbs_error.h"
#include	"log.h"
#include	"net_connect.h"
#include	"rpp.h"
#include	"dis.h"
#include	"dis_init.h"

static char ident[] = "@(#) $RCSfile: mom_server.c,v $ $Revision: 2.0 $";


/* Global Data Items */

extern	unsigned int	default_server_port;
extern	char		mom_host[];
extern	char		*path_jobs;
extern	char		*path_home;
extern	int		pbs_errno;
extern	unsigned int	pbs_mom_port;
extern	unsigned int	pbs_rm_port;
extern	unsigned int	pbs_tm_port;
extern	time_t		time_now;

/*
 * Tree search generalized from Knuth (6.2.2) Algorithm T just like
 * the AT&T man page says.
 *
 * The node_t structure is for internal use only, lint doesn't grok it.
 *
 * Written by reading the System V Interface Definition, not the code.
 *
 * Totally public domain.
 */
/*LINTLIBRARY*/

/*
**	Modified by Tom Proett <proett@nas.nasa.gov> for PBS.
*/

typedef struct node_t {
	u_long		key;
	struct node_t	*left, *right;
} node;
node		*okclients = NULL;	/* tree of ip addrs */

/* find value in tree, return 1 if found, 0 if not */
int
tfind(key, rootp)
	const u_long 	key;		/* key to be located */
	node		**rootp;	/* address of tree root */
{
	register node	*q;

	if (rootp == (struct node_t **)0)
		return 0;
	while (*rootp != (struct node_t *)0) {	/* Knuth's T1: */
		if (key == (*rootp)->key)	/* T2: */
			return 1;		/* we found it! */
		rootp = (key < (*rootp)->key) ?
			&(*rootp)->left :	/* T3: follow left branch */
			&(*rootp)->right;	/* T4: follow right branch */
	}
	return 0;
}

void
tinsert(key, rootp)
	const u_long 	key;		/* key to be located */
	node		**rootp;	/* address of tree root */
{
	register node	*q;

	if (rootp == (struct node_t **)0)
		return;
	while (*rootp != (struct node_t *)0) {	/* Knuth's T1: */
		if (key == (*rootp)->key)	/* T2: */
			return;			/* we found it! */
		rootp = (key < (*rootp)->key) ?
			&(*rootp)->left :	/* T3: follow left branch */
			&(*rootp)->right;	/* T4: follow right branch */
	}
	q = (node *) malloc(sizeof(node));	/* T5: key not found */
	if (q != (struct node_t *)0) {		/* make new node */
		*rootp = q;			/* link new node to old */
		q->key = key;			/* initialize new node */
		q->left = q->right = (struct node_t *)0;
	}
	return;
}

void
tfree(rootp)
	node		**rootp;
{
	if (rootp == NULL || *rootp == NULL)
		return;
	tfree(&(*rootp)->left);
	tfree(&(*rootp)->right);
	free(*rootp);
	*rootp = NULL;
}

/*
**	Start a standard inter-server message.
*/
int
is_compose(stream, command)
     int	stream;
     int	command;
{
	int	ret;

	if (stream < 0)
		return DIS_EOF;
	DIS_rpp_reset();

	ret = diswsi(stream, IS_PROTOCOL);
	if (ret != DIS_SUCCESS)
		goto done;
	ret = diswsi(stream, IS_PROTOCOL_VER);
	if (ret != DIS_SUCCESS)
		goto done;
	ret = diswsi(stream, command);
	if (ret != DIS_SUCCESS)
		goto done;
	return DIS_SUCCESS;

 done:
	DBPRT(("is_compose: send error %s\n", dis_emsg[ret]))
	return ret;
}

/*
**	Input is coming from another server over a DIS rpp stream.
**	Read the stream to get a Inter-Server request.
*/
void
is_request(stream, version)
     int	stream;
     int	version;
{
	static	char		id[] = "is_request";
	int			command = 0;
	int			ret = DIS_SUCCESS;
	int			i;
	u_long			ipaddr;
	short			port;
	struct	sockaddr_in	*addr;
	void			init_addrs();

	DBPRT(("%s: stream %d version %d\n", id, stream, version))
	if (version != IS_PROTOCOL_VER) {
		sprintf(log_buffer, "protocol version %d unknown", version);
		log_err(-1, id, log_buffer);
		rpp_close(stream);
		return;
	}

	/* check that machine is okay to be a server */
	addr = rpp_getaddr(stream);
	port = ntohs((unsigned short)addr->sin_port);
	ipaddr = ntohl(addr->sin_addr.s_addr);
	if (port >= IPPORT_RESERVED || !tfind(ipaddr, &okclients)) {
		sprintf(log_buffer, "bad connect from %s",
			netaddr(addr));
		log_err(-1, id, log_buffer);
		rpp_close(stream);
		return;
	}

	command = disrsi(stream, &ret);
	if (ret != DIS_SUCCESS)
		goto err;

	switch (command) {

	case IS_NULL:		/* a ping from the server */
		DBPRT(("%s: IS_NULL\n", id))
		break;

	case IS_HELLO:		/* server wants a return ping */
		DBPRT(("%s: IS_HELLO\n", id))
		is_compose(stream, IS_HELLO);
		rpp_flush(stream);
		break;

	case IS_CLUSTER_ADDRS:
		DBPRT(("%s: IS_CLUSTER_ADDRS\n", id))
		for (;;) {
			ipaddr = disrul(stream, &ret);
			if (ret != DIS_SUCCESS)
				break;
			DBPRT(("%s:\t%ld.%ld.%ld.%ld\n", id,
				(ipaddr & 0xff000000) >> 24,
				(ipaddr & 0x00ff0000) >> 16,
				(ipaddr & 0x0000ff00) >> 8,
				(ipaddr & 0x000000ff) ))
			tinsert(ipaddr, &okclients);
		}
		if (ret != DIS_EOD)
			goto err;
		break;

	default:
		sprintf(log_buffer, "unknown command %d sent", command);
		log_err(-1, id, log_buffer);
		goto err;
	}

	rpp_eom(stream);
	return;

 err:
	/*
	** We come here if we got a DIS read error or a protocol
	** element is missing.
	*/
	sprintf(log_buffer, "%s from %s", dis_emsg[ret], netaddr(addr));
	log_err(-1, id, log_buffer);
	rpp_close(stream);

	return;
}
